import os
import unittest
from cfitall import utils


class TestAddKeys(unittest.TestCase):
    def test_add_keys_single(self):
        new_dict = utils.add_keys({}, ['foo'])
        self.assertEqual(new_dict, {'foo': None})

    def test_add_keys_single_with_value(self):
        new_dict = utils.add_keys({}, ['foo'], 'bar')
        self.assertEqual(new_dict, {'foo': 'bar'})

    def test_add_keys_nested(self):
        new_dict = utils.add_keys({}, ['foo', 'bar', 'baz'])
        self.assertEqual(new_dict, {'foo': {'bar': {'baz': None}}})

    def test_add_keys_nested_with_value(self):
        new_dict = utils.add_keys({}, ['foo', 'bar', 'baz'], 'bat')
        self.assertEqual(new_dict, {'foo': {'bar': {'baz': 'bat'}}})

    def test_add_keys_nested_with_value_and_dict(self):
        start_dict = {'asdf': {'1234': 'rewq'}}
        new_dict = utils.add_keys(start_dict, ['foo', 'bar', 'baz'], 'bat')
        self.assertEqual(new_dict, {
            'asdf': {'1234': 'rewq'},
            'foo': {'bar': {'baz': 'bat'}}
        })


class TestExpandFlatten(unittest.TestCase):
    def test_expand_flattened_path(self):
        expanded = utils.expand_flattened_path('asdf.fdsa.qwer.rewq', 'foo')
        self.assertEqual(expanded, {'asdf': {'fdsa': {'qwer': {'rewq': 'foo'}}}})

    def test_expand_flattened_dict(self):
        expanded = utils.expand_flattened_dict({'asdf.fdsa.qwer.rewq': 'foo'})
        self.assertEqual(expanded, {'asdf': {'fdsa': {'qwer': {'rewq': 'foo'}}}})

    def test_flatten_dict(self):
        flattened = utils.flatten_dict({'asdf': {'fdsa': {'qwer': {'rewq': 'foo'}}}})
        self.assertEqual(flattened, {'asdf.fdsa.qwer.rewq': 'foo'})


class TestExtractFindMerge(unittest.TestCase):
    def test_find_keys(self):
        keys = utils.find_keys({'asdf': {'fdsa': {'qwer': {'rewq': 'foo'}}}}, 'foo')
        self.assertEqual(keys, ['asdf', 'fdsa', 'qwer', 'rewq'])

    def test_extract_values(self):
        extracted = utils.extract_values({'asdf': {'fdsa': {'qwer': {'rewq': 'foo'}}}})
        self.assertEqual(extracted, ['foo'])

    def test_merge_dicts(self):
        srcdict = {'asdf': 'fdsa', 'qwer': {'werq': 'poiu'}}
        destdict = {'lkjh': 'zxcv', 'asdf': 1234}
        merged = utils.merge_dicts(srcdict, destdict)
        self.assertEqual(merged, {
            'asdf': 'fdsa',
            'qwer': {'werq': 'poiu'},
            'lkjh': 'zxcv'
        })


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from cfitall.config import ConfigManager


class TestConfigManager(unittest.TestCase):
    def setUp(self):
        super().setUp()
        os.environ['CFITALL__GLOBAL__NAME'] = 'cfitall'
        os.environ['CFITALL__GLOBAL__PATH'] = '/Users/wryfi,/Users/wryfi/tmp'
        os.environ['CFITALL__FOO__BANG'] = 'WHAMMY!'

    def tearDown(self):
        super().tearDown()
        del os.environ['CFITALL__GLOBAL__NAME']
        del os.environ['CFITALL__GLOBAL__PATH']
        del os.environ['CFITALL__FOO__BANG']

    def test_create_object(self):
        _cf = ConfigManager('cfitall')

    def test_environment_variables(self):
        cf = ConfigManager('cfitall')
        self.assertEqual(cf.get('global.name'), 'cfitall')
        self.assertEqual(cf.get('global.path', list), ['/Users/wryfi', '/Users/wryfi/tmp'])
        self.assertEqual(cf.get('foo.bang'), 'WHAMMY!')

    def test_config_file(self):
        cf = ConfigManager('cfitall')
        cf.add_config_path(os.path.dirname(os.path.abspath(__file__)))
        cf.read_config()
        self.assertEqual(cf.get('global.name'), 'cfitall')
        self.assertEqual(cf.get('global.path', list), ['/Users/wryfi', '/Users/wryfi/tmp'])
        self.assertEqual(cf.get('foo.bang'), 'WHAMMY!')
        self.assertEqual(cf.get('global.authors', list), ['john doe', 'jane deer'])
        self.assertEqual(cf.get('search', list), ['/Users/wryfi/foo', '/Users/wryfi/bar'])
        self.assertEqual(cf.get('foo.bar'), 'baz')

    def test_config_order(self):
        cf = ConfigManager('cfitall')
        cf.set_default('global.authors', ['michael scott', 'jim halpert'])
        self.assertEqual(cf.get('global.authors', list), ['michael scott', 'jim halpert'])
        cf.add_config_path(os.path.dirname(os.path.abspath(__file__)))
        cf.read_config()
        self.assertEqual(cf.get('global.authors', list), ['john doe', 'jane deer'])
        os.environ['CFITALL__GLOBAL__AUTHORS'] = 'holly flax,robert california'
        self.assertEqual(cf.get('global.authors', list), ['holly flax', 'robert california'])
        cf.set('global.authors', ['dwight schrute', 'pam beesly'])
        self.assertEqual(cf.get('global.authors', list), ['dwight schrute', 'pam beesly'])


if __name__ == '__main__':
    unittest.main()



from decimal import Decimal
import json
import re
import os
import yaml

from cfitall import utils


class ConfigManager(object):
    def __init__(self, name, env_prefix=None, env_path_sep='__', env_value_split=True, defaults={}):
        """
        The configuration registry holds configuration data from different sources
        and reconciles it for retrieval.

        :param str name: name of registry (cannot contain env_separator string)
        :param str env_prefix: prefix for environment variables (defaults to uppercase name)
        :param str env_path_sep: string for separating config hierarchies in env vars (default '__')
        :param bool env_value_split: if True, then environment variable values are split on commas into a list
        :param dict defaults: dictionary of default configuration settings
        """
        self.name = name
        self.config_file = None
        self.config_path = []
        self.values = {'super': {}, 'cli': {}, 'cfgfile': {}, 'defaults': defaults}
        self.env_path_sep = env_path_sep
        self.env_value_split = env_value_split
        if env_prefix:
            self.env_prefix = env_prefix.upper()
        else:
            self.env_prefix = self.name.upper()

    @property
    def config_keys(self):
        """
        Returns a list of configuration keys as dotted paths, for
        use with the get() or set() methods.

        :return: list of configuration keys as dotted paths
        :rtype: list
        """
        config_keys = [key for key, value in self.flattened.items()]
        return sorted(config_keys)

    @property
    def dict(self):
        """
        Returns a dict of merged configuration data

        :return: merged dictionary of configuration data
        :rtype: dict
        """
        return self._merge_configs()

    @property
    def env_vars(self):
        """
        Returns a list of environment variables known from config files and defaults

        :return: list of environment variables that will be read
        :rtype: list
        """
        prefix = self.env_prefix + self.env_path_sep
        keys = [key.upper() for key, value in self.flattened.items()]
        keys = [prefix + key.replace('.', self.env_path_sep) for key in keys]
        return sorted(keys)

    @property
    def flattened(self):
        """
        Returns a "flattened" dictionary of merged config values,
        condensing hierarchies into dotted paths and returning simple
        key-value pairs.

        :return: flattened dictionary of merged config values
        :rtype: dict
        """
        return utils.flatten_dict(self.dict)

    @property
    def json(self):
        return json.dumps(self.dict, indent=4, sort_keys=True)

    @property
    def yaml(self):
        return yaml.dump(self.dict)

    def add_config_path(self, path):
        """
        Adds a path to search for a configuration file.  Currently
        limited to the local filesystem, s3 integration envisioned.

        :param path: filesystem path to search for config files
        :return: None
        """
        self.config_path.append(path)

    def get(self, config_key, rtype=None):
        """
        Get a configuration value by its dotted path key.  There
        must be an exact match for the value you request.

        :param config_key: dotted path key in the config registry
        :param rtype: requested return type (list, str, int, Decimal, float)
        :return: value from config registry corresponding to key
        """
        try:
            value = self.flattened[config_key]
        except KeyError:
            return None
        value_type = type(value)
        if rtype == list:
            return list(value)
        if rtype == str:
            if value_type == list:
                return ','.join(value)
            return str(value)
        elif rtype == int:
            return int(value)
        elif rtype == Decimal:
            return Decimal(value)
        elif rtype == float:
            return float(value)
        return value

    def set(self, config_key, value):
        """
        Explicitly set a configuration key via dotted key path.
        Configurations set this way take precedence over all other
        configuration sources.

        :param config_key: dotted path key to set
        :param value: value to set
        """
        flat_dict = {config_key: value}
        expanded = utils.expand_flattened_dict(flat_dict)
        utils.merge_dicts(expanded, self.values['super'])

    def set_default(self, config_key, value):
        """
        Set a default value in the registry via dotted key path.
        Configurations set this way are the first to be overriden by other
        configuration sources.

        :param config_key: dotted path key to set
        :param value: value to set
        :return:
        """
        flat_dict = {config_key: value}
        expanded = utils.expand_flattened_dict(flat_dict)
        utils.merge_dicts(expanded, self.values['defaults'])

    def read_config(self):
        """
        Search through the available paths in config_path and read the first
        suitable configuration file found.

        :return: True if a configuration file was read, else False
        :rtype: bool
        """
        for path in self.config_path:
            if os.path.isdir(path):
                for file in os.listdir(path):
                    if re.match('{}.json'.format(self.name.lower()), file):
                        self.config_file = os.path.join(path, file)
                        self._read_json_file(self.config_file)
                        return True
                    if re.match('{}.ya*ml'.format(self.name.lower()), file):
                        self.config_file = os.path.join(path, file)
                        self._read_yaml_file(self.config_file)
                        return True
        return False

    def _read_yaml_file(self, path):
        """
        Opens path as a yaml file and attempts to safely load it into the
        cfgfile value dictionary.

        :param path: path to yaml file
        """
        with open(path, 'r') as fp:
            data = yaml.safe_load(fp.read())
        for key, value in data.items():
            self.values['cfgfile'][key.lower()] = value

    def _read_json_file(self, path):
        """
        Opens path as a json file and attempts to load it into the
        cfgfile value dictionary.

        :param path: path to json file
        """
        with open(path, 'r') as fp:
            data = json.loads(fp.read())
        for key, value in data.items():
            self.values['cfgfile'][key.lower()] = value

    def _read_environment(self):
        """
        Reads all environment variables beginning with env_prefix and loads
        them into a dictionary using env_separator as a hierarchical path
        separator.

        :return: config dictionary read from environment variables
        :rtype: dict
        """
        output = {}
        prefix = self.env_prefix + self.env_path_sep
        for key, value in os.environ.items():
            if key.startswith(prefix):
                key = key.replace(prefix, '', 1).lower()
                if isinstance(value, str) and self.env_value_split:
                    if re.match(r'.*,(.*,)*.*', value):
                        value = value.split(',')
                output[key] = value
        return utils.expand_flattened_dict(output, separator=self.env_path_sep)

    def _merge_configs(self):
        """
        Merges all of the configuration data together in the appropriate order.

        :return: merged configuration data
        :rtype: dict
        """
        config = utils.merge_dicts(self.values['defaults'], {})
        config = utils.merge_dicts(self.values['cfgfile'], config)
        config = utils.merge_dicts(self._read_environment(), config)
        config = utils.merge_dicts(self.values['super'], config)
        return config



"""
This module is full of little recursive functions that help with
converting string-path notations into nested dicts, and vice
versa. A string-path, or "flattened" dict might look like this:

    {'foo.bar.bat': 'asdfhjklkjhfdsa'}

As an "expanded" or "nested" dict, the same data would be:

    {'foo': {'bar': {'bat': 'lkjhgfdsasdfghjkl'}}}
"""

from collections import Mapping
import re


def add_keys(destdict, srclist, value=None):
    """
    Nests keys from srclist into destdict, with optional value set on the final key.

    :param dict destdict: the dict to update with values from srclist
    :param list srclist: list of keys to add to destdict
    :param value: final value to set
    :return: destination dictionary
    :rtype: dict
    """
    if len(srclist) > 1:
        destdict[srclist[0]] = {}
        destdict[srclist[0]] = destdict.get(srclist[0], {})
        add_keys(destdict[srclist[0]], srclist[1:], value)
    else:
        destdict[srclist[0]] = value
    return destdict


def expand_flattened_path(flattened_path, value=None, separator='.'):
    """
    Expands a dotted path into a nested dict; if value is set, the
    final key in the path will be set to value.

    :param str flattened_path: flattened path to expand
    :param str separator: character(s) separating path components
    :param value: set final key to this value
    :return: nested dictionary
    :rtype: dict
    """
    split_list = flattened_path.split(separator)
    return add_keys({}, split_list, value)


def find_keys(searchdict, searchvalue):
    """
    Returns a list of all the parent keys for searchvalue in searchdict.

    :param dict searchdict: dictionary to search
    :param str searchvalue: value to search dictionary keys for
    :return: list of dictionary keys matching searchvalue
    :rtype: list
    """
    for key, value in searchdict.items():
        if isinstance(value, Mapping):
            path = find_keys(value, searchvalue)
            if path:
                return [key] + path
        elif value == searchvalue:
            return [key]


def extract_values(searchdict):
    """
    Returns a list of all non-mapping values extracted from (nested) searchdict

    :param dict searchdict: dictionary to extract values from
    :param list values: list of existing values to pass in (used to recurse)
    :return: list of values
    :rtype: list
    """
    values = []
    for key, value in searchdict.items():
        if isinstance(value, Mapping):
            for value in extract_values(value):
                values.append(value)
        elif value not in values:
            values.append(value)
    return values


def flatten_dict(nested, separator='.'):
    """
    Converts a nested dict into a flattened dict, with keys in dotted path notation

    :param dict nested: nested dict to flatten
    :param str separator: charactor used to separate paths in flattened dict
    :return: flattened dict
    :rtype: dict
    """
    flattened = {}
    values = extract_values(nested)
    for value in values:
        keys = find_keys(nested, value)
        flattened_path = separator.join(keys)
        flattened[flattened_path] = value
    return flattened


def merge_dicts(source, destination):
    """
    Performs a deep merge of two nested dicts by expanding all Mapping objects
    until they reach a non-mapping value (e.g. a list, string, int, etc.) and
    copying these from the source to the destination.

    :param dict source: the source dictionary to copy values from
    :param dict destination: the dictionary to update with values from source
    :return: destination
    :rtype: dict
    """
    for key, value in source.items():
        key = key.lower() if isinstance(key, str) else key
        if isinstance(value, Mapping):
            node = destination.setdefault(key, {})
            merge_dicts(value, node)
        else:
            destination[key] = value
    return destination


def expand_flattened_dict(flattened, separator='.'):
    """
    Expands a flattened dict into a nested dict.

    :param dict flattened: the flattened dict to expand
    :param str separator: character used for separating paths in flattened dict
    :return: nested dict
    :rtype: dict
    """
    merged = {}
    for key, value in flattened.items():
        expanded = expand_flattened_path(key, value=value, separator=separator)
        merged = merge_dicts(merged, expanded)
    return merged
